from typing import Optional
from skyplane.config_paths import config_path, ibmcloud_config_path
from skyplane.config import SkyplaneConfig
from skyplane.utils import imports

PUBLIC_ENDPOINT = "https://s3.{}.cloud-object-storage.appdomain.cloud"
PRIVATE_ENDPOINT = "https://s3.private.{}.cloud-object-storage.appdomain.cloud"
DIRECT_ENDPOINT = "https://s3.direct.{}.cloud-object-storage.appdomain.cloud"

OBJ_REQ_RETRIES = 5
CONN_READ_TIMEOUT = 10
VPC_API_VERSION = "2021-09-21"


class IBMCloudAuthentication:
    def __init__(self, config: Optional[SkyplaneConfig] = None):
        """Loads IBM Cloud authentication details. If no access key is provided, it will try to load credentials using boto3"""
        if not config is None:
            self.config = config
        else:
            self.config = SkyplaneConfig.load_config(config_path)

        self.user_agent = self.config.ibmcloud_useragent if self.config.ibmcloud_useragent is not None else "skyplane-ibm"
        self._ibmcloud_resource_group_id = self.config.ibmcloud_resource_group_id

        if self.config.ibmcloud_access_id and self.config.ibmcloud_secret_key:
            self._access_key = self.config.ibmcloud_access_id
            self._secret_key = self.config.ibmcloud_secret_key
        if self.config.ibmcloud_iam_key:
            self._iam_key = self.config.ibmcloud_iam_key

    @imports.inject("ibm_cloud_sdk_core", "ibm_cloud_sdk_core.authenticators", pip_extra="ibmcloud")
    def get_iam_authenticator(ibm_cloud_sdk_core, self):
        return ibm_cloud_sdk_core.authenticators.IAMAuthenticator(self.config.ibmcloud_iam_key, url=self.config.ibmcloud_iam_endpoint)

    def get_ibmcloud_endpoint(self, region, compute_backend="public"):
        if region is not None:
            endpoint = PUBLIC_ENDPOINT.format(region)

            if compute_backend == "ibm_vpc":
                endpoint = DIRECT_ENDPOINT.format(region)

            return endpoint

    @imports.inject("ibm_cloud_sdk_core", "ibm_vpc", pip_extra="ibmcloud")
    def save_region_config(ibm_cloud_sdk_core, ibm_vpc, self, config: SkyplaneConfig):
        if not config.ibmcloud_enabled:
            self.clear_region_config()
            return
        with ibmcloud_config_path.open("w") as f:
            region_list = []
            authenticator = ibm_cloud_sdk_core.authenticators.IAMAuthenticator(config.ibmcloud_iam_key, url=config.ibmcloud_iam_endpoint)
            ibm_vpc_client = ibm_vpc.VpcV1(VPC_API_VERSION, authenticator=authenticator)
            res = ibm_vpc_client.list_regions()
            for region in res.result["regions"]:
                if region["status"] == "available":
                    zones = ibm_vpc_client.list_region_zones(region["name"])
                    for zone in zones.result["zones"]:
                        if zone["status"] == "available":
                            region_list.append("{},{},{},{}".format(zone["region"]["name"], region["href"], zone["name"], zone["href"]))
            f.write("\n".join(region_list))

    def clear_region_config(self):
        with ibmcloud_config_path.open("w") as f:
            f.write("")

    @staticmethod
    def get_region_config():
        try:
            f = open(ibmcloud_config_path, "r")
        except FileNotFoundError:
            return []
        region_list = {}
        for region in f.read().split("\n"):
            line = region.split(",")
            if line[0] not in region_list:
                region_list[line[0]] = {}
                region_list[line[0]]["zones"] = []
            region_list[line[0]]["href"] = line[1]
            region_list[line[0]]["zones"].append({"zone_name": line[2], "zone_href": line[3]})

        return region_list

    @property
    def access_key(self):
        return self._access_key

    @property
    def iam_api_key(self):
        return self._iam_key

    @property
    def iam_endpoint(self):
        return self.config.ibmcloud_iam_endpoint

    @property
    def ibmcloud_resource_group_id(self):
        return self._ibmcloud_resource_group_id

    @property
    def secret_key(self):
        return self._secret_key

    def enabled(self):
        return self.config.ibmcloud_enabled

    @imports.inject("ibm_boto3", pip_extra="ibmcloud")
    def get_boto3_session(ibm_boto3, self, cos_region: Optional[str] = None):
        return ibm_boto3.Session(aws_access_key_id=self.access_key, aws_secret_access_key=self.secret_key)

    def get_boto3_resource(self, service_name, cos_region=None):
        return self.get_boto3_session().resource(service_name, region_name=cos_region)

    @imports.inject("ibm_boto3", "ibm_botocore", pip_extra="ibmcloud")
    def get_boto3_client(ibm_boto3, ibm_botocore, self, service_name, cos_region=None):
        client_config = ibm_botocore.client.Config(
            max_pool_connections=128,
            user_agent_extra=self.user_agent,
            connect_timeout=CONN_READ_TIMEOUT,
            read_timeout=CONN_READ_TIMEOUT,
            retries={"max_attempts": OBJ_REQ_RETRIES},
        )

        if cos_region is None:
            return self.get_boto3_session().client(service_name, config=client_config)
        else:
            return self.get_boto3_session().client(service_name, endpoint_url=self.get_ibmcloud_endpoint(cos_region), config=client_config)



# TODO: migrate to programmable gateways
# from skyplane.gateway.gateway_sender import GatewaySender
#
#
# class GatewayOnPrem(GatewaySender):
#    def start_workers(self):
#        # Assert no other prallel internet connections are active on the node
#        # Else this could lead to a noisy neighbor problem
#        assert len(psutil.net_connections()) < 5, "Cannot start workers when other workers are running"
#        for ip, num_connections in self.outgoing_ports.items():
#            for i in range(num_connections):
#                p = Process(target=self.worker_loop, args=(i, ip))
#                p.start()
#                self.processes.append(p)
#



import os
from pathlib import Path

from skyplane import exceptions as skyplane_exceptions
from skyplane.compute.aws.aws_auth import AWSAuthentication
from skyplane.compute.server import key_root
from skyplane.utils import logger


class AWSKeyManager:
    """Stores SSH keys for access to AWS VMs."""

    def __init__(self, auth: AWSAuthentication, local_key_dir: Path = key_root / "aws"):
        self.auth = auth
        self.local_key_dir = local_key_dir

    def key_exists_aws(self, aws_region: str, key_name: str) -> bool:
        """Checks if a key exists in AWS."""
        ec2_client = self.auth.get_boto3_client("ec2", aws_region)
        return key_name in set(p["KeyName"] for p in ec2_client.describe_key_pairs()["KeyPairs"])

    def key_exists_local(self, key_name: str) -> bool:
        """Checks if a key exists locally."""
        return (self.local_key_dir / f"{key_name}.pem").exists()

    def make_key(self, aws_region: str, key_name: str) -> Path:
        """Creates a key in AWS and stores it locally."""
        if self.key_exists_aws(aws_region, key_name):
            logger.error(f"Key {key_name} already exists in AWS region {aws_region}")
            raise skyplane_exceptions.PermissionsException(
                f"Key {key_name} already exists in AWS region {aws_region}, please delete it first or use a different key name."
            )
        if self.key_exists_local(key_name):
            logger.error(f"Key {key_name} already exists locally")
            raise skyplane_exceptions.PermissionsException(
                f"Key {key_name} already exists locally, please delete it first or use a different key name."
            )
        ec2 = self.auth.get_boto3_resource("ec2", aws_region)
        local_key_file = self.local_key_dir / f"{key_name}.pem"
        local_key_file.parent.mkdir(parents=True, exist_ok=True)
        logger.fs.debug(f"[AWS] Creating keypair {key_name} in {aws_region}")
        key_pair = ec2.create_key_pair(KeyName=key_name, KeyType="rsa")
        with local_key_file.open("w") as f:
            key_str = key_pair.key_material
            if not key_str.endswith("\n"):
                key_str += "\n"
            f.write(key_str)
        os.chmod(local_key_file, 0o600)
        return local_key_file

    def delete_key(self, aws_region: str, key_name: str):
        """Deletes a key from AWS and locally."""
        if self.key_exists_aws(aws_region, key_name):
            ec2 = self.auth.get_boto3_resource("ec2", aws_region)
            logger.fs.debug(f"[AWS] Deleting keypair {key_name} in {aws_region}")
            ec2.KeyPair(key_name).delete()
        if self.key_exists_local(key_name):
            (self.local_key_dir / f"{key_name}.pem").unlink()

    def get_key(self, key_name: str) -> Path:
        """Returns path to local keyfile."""
        return self.local_key_dir / f"{key_name}.pem"

    def ensure_key_exists(self, aws_region: str, key_name: str, delete_remote: bool = True) -> Path:
        """Ensures that a key exists in AWS and locally, creating it if necessary. Raise an exception if it's on AWS and not locally."""
        local_exists, remote_exists = self.key_exists_local(key_name), self.key_exists_aws(aws_region, key_name)
        if local_exists and remote_exists:
            return self.get_key(key_name)
        elif not local_exists and not remote_exists:
            return self.make_key(aws_region, key_name)
        elif local_exists and not remote_exists:
            local_key_path = self.get_key(key_name)
            logger.warning(f"Key {key_name} exists locally but not in AWS region {aws_region}. Moving the local key {local_key_path}.bak")
            local_key_path.rename(local_key_path.with_suffix(".pem.bak"))
            return self.make_key(aws_region, key_name)
        else:
            if delete_remote:
                logger.warning(f"Key {key_name} exists in AWS region {aws_region} but not locally. Deleting the remote key.")
                self.delete_key(aws_region, key_name)
                return self.make_key(aws_region, key_name)
            else:
                raise skyplane_exceptions.PermissionsException(
                    f"Key {key_name} exists in AWS region {aws_region} but not locally. Please delete the key from AWS or move it locally."
                )



import socket
from dataclasses import asdict, dataclass
from enum import Enum, auto
from functools import total_ordering

from typing import Dict, Optional


@dataclass
class Chunk:
    """A Chunk is a contiguous piece of a file (a file may be one or more chunks)."""

    src_key: str  # human readable path where object is stored
    dest_key: str  # human readable path where object is stored
    chunk_id: str
    chunk_length_bytes: int
    partition_id: Optional[str] = None  # for broadcast
    mime_type: Optional[str] = None

    # checksum
    md5_hash: Optional[bytes] = None  # 128 bits

    # multi-part upload/download info
    multi_part: Optional[bool] = False
    file_offset_bytes: Optional[int] = None
    part_number: Optional[int] = None
    upload_id: Optional[str] = None  # TODO: for broadcast, this is not used

    def to_wire_header(self, n_chunks_left_on_socket: int, wire_length: int, raw_wire_length: int, is_compressed: bool = False):
        return WireProtocolHeader(
            chunk_id=self.chunk_id,
            data_len=wire_length,
            raw_data_len=raw_wire_length,
            is_compressed=is_compressed,
            n_chunks_left_on_socket=n_chunks_left_on_socket,
        )

    def as_dict(self):
        return asdict(self)

    @staticmethod
    def from_dict(d: Dict):
        return Chunk(**d)


# TODO: remove ChunkRequest abstraction (only need chunks)
@dataclass
class ChunkRequest:
    """A ChunkRequest stores all local state in the Gateway pertaining to a ChunkRequest."""

    chunk: Chunk
    src_region: Optional[str] = None
    dst_region: Optional[str] = None
    src_type: Optional[str] = None  # enum of {"object_store", "random", "read_local"}
    dst_type: Optional[str] = None  # enum of {"object_store", "save_local"}
    src_random_size_mb: Optional[int] = None
    src_object_store_bucket: Optional[str] = None
    dst_object_store_bucket: Optional[str] = None

    def __post_init__(self):
        if self.src_type == "object_store":
            assert self.src_object_store_bucket is not None
        elif self.src_type == "random":
            assert self.src_random_size_mb is not None
        if self.dst_type == "object_store":
            assert self.dst_object_store_bucket is not None

    def as_dict(self):
        dict_out = asdict(self)
        dict_out["chunk"] = self.chunk.as_dict()
        return dict_out

    @staticmethod
    def from_dict(in_dict: Dict):
        # in_dict["chunk"] = Chunk.from_dict(in_dict)
        return ChunkRequest(**{"chunk": Chunk.from_dict(in_dict)})


@total_ordering
class ChunkState(Enum):
    registered = auto()
    in_progress = auto()
    failed = auto()
    queued = auto()
    complete = auto()

    @staticmethod
    def from_str(s: str):
        return ChunkState[s.lower()]

    def __lt__(self, other):
        return self.value < other.value


@dataclass
class WireProtocolHeader:
    """Lightweight wire protocol header for chunk transfers along socket."""

    chunk_id: str  # 128bit UUID
    data_len: int  # long
    raw_data_len: int  # long (uncompressed, unecrypted)
    is_compressed: bool  # char
    n_chunks_left_on_socket: int  # long

    @staticmethod
    def magic_hex():
        return 0x534B595F4C41524B  # "SKY_LARK"

    @staticmethod
    def protocol_version():
        # v1 = base protocol
        # v2 = compression
        # v3 = uuid chunk_id
        return 3

    @staticmethod
    def length_bytes():
        # magic (8) + protocol_version (4) + chunk_id (16) + data_len (8) + raw_data_len(8) + is_compressed (1) + n_chunks_left_on_socket (8)
        return 8 + 4 + 16 + 8 + 8 + 1 + 8

    @staticmethod
    def from_bytes(data: bytes):
        assert len(data) == WireProtocolHeader.length_bytes(), f"{len(data)} != {WireProtocolHeader.length_bytes()}"
        magic = int.from_bytes(data[:8], byteorder="big")
        if magic != WireProtocolHeader.magic_hex():
            raise ValueError(f"Invalid magic number, got {magic:x} but expected {WireProtocolHeader.magic_hex():x}")
        version = int.from_bytes(data[8:12], byteorder="big")
        if version != WireProtocolHeader.protocol_version():
            raise ValueError(f"Invalid protocol version, got {version} but expected {WireProtocolHeader.protocol_version()}")
        chunk_id = data[12:28].hex()
        chunk_len = int.from_bytes(data[28:36], byteorder="big")
        raw_chunk_len = int.from_bytes(data[36:44], byteorder="big")
        is_compressed = bool(int.from_bytes(data[44:45], byteorder="big"))
        n_chunks_left_on_socket = int.from_bytes(data[45:53], byteorder="big")
        return WireProtocolHeader(
            chunk_id=chunk_id,
            data_len=chunk_len,
            raw_data_len=raw_chunk_len,
            is_compressed=is_compressed,
            n_chunks_left_on_socket=n_chunks_left_on_socket,
        )

    def to_bytes(self):
        out_bytes = b""
        out_bytes += self.magic_hex().to_bytes(8, byteorder="big")
        out_bytes += self.protocol_version().to_bytes(4, byteorder="big")
        chunk_id_bytes = bytes.fromhex(self.chunk_id)
        assert len(chunk_id_bytes) == 16
        out_bytes += chunk_id_bytes
        out_bytes += self.data_len.to_bytes(8, byteorder="big")
        out_bytes += self.raw_data_len.to_bytes(8, byteorder="big")
        out_bytes += self.is_compressed.to_bytes(1, byteorder="big")
        out_bytes += self.n_chunks_left_on_socket.to_bytes(8, byteorder="big")
        assert len(out_bytes) == WireProtocolHeader.length_bytes(), f"{len(out_bytes)} != {WireProtocolHeader.length_bytes()}"
        return out_bytes

    @staticmethod
    def from_socket(sock: socket.socket):
        num_bytes = WireProtocolHeader.length_bytes()
        header_bytes = b""
        while len(header_bytes) < num_bytes:
            header_bytes += sock.recv(num_bytes - len(header_bytes))
        assert len(header_bytes) == num_bytes, f"{len(header_bytes)} != {num_bytes}"
        return WireProtocolHeader.from_bytes(header_bytes)

    def to_socket(self, sock: socket.socket):
        assert sock.sendall(self.to_bytes()) is None



import paramiko
import logging
import os

logger = logging.getLogger(__name__)


class SSHClient:
    def __init__(self, ip_address, ssh_credentials):
        self.ip_address = ip_address
        self.ssh_credentials = ssh_credentials
        self.ssh_client = None

        if "key_filename" in self.ssh_credentials:
            fpath = os.path.expanduser(self.ssh_credentials["key_filename"])
            if not os.path.exists(fpath):
                raise Exception(f"Private key file {fpath} doesn't exist")
            self.ssh_credentials["key_filename"] = fpath

    def close(self):
        """
        Closes the SSH client connection
        """
        self.ssh_client.close()
        self.ssh_client = None

    def create_client(self, timeout=2):
        """
        Crate the SSH client connection
        """
        try:
            self.ssh_client = paramiko.SSHClient()
            self.ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

            user = self.ssh_credentials.get("username")
            password = self.ssh_credentials.get("password")
            pkey = None

            if self.ssh_credentials.get("key_filename"):
                with open(self.ssh_credentials["key_filename"]) as f:
                    pkey = paramiko.RSAKey.from_private_key(f)

            self.ssh_client.connect(
                self.ip_address,
                username=user,
                password=password,
                pkey=pkey,
                timeout=timeout,
                banner_timeout=200,
                allow_agent=False,
                look_for_keys=False,
            )

            logger.debug(f"{self.ip_address} ssh client created")
        except Exception as e:
            pk = self.ssh_credentials.get("key_filename")
            if pk and str(e) == "Authentication failed.":
                raise Exception(f"Private key {pk} is not valid")
            raise e

        return self.ssh_client

    def exec_command(self, cmd, timeout=None, run_async=False):
        if not self.ip_address or self.ip_address == "0.0.0.0":
            raise Exception("Invalid IP Address")

        if self.ssh_client is None:
            self.ssh_client = self.create_client()

        try:
            return self.ssh_client.exec_command(cmd, timeout=timeout)
        except Exception:
            # Normally this is a timeout exception
            self.ssh_client = self.create_client()
            return self.ssh_client.exec_command(cmd, timeout=timeout)

    def run_remote_command(self, cmd, timeout=None, run_async=False):
        """
        Executa a command
        param: timeout: execution timeout
        param: run_async: do not wait for command completion
        """
        if not self.ip_address or self.ip_address == "0.0.0.0":
            raise Exception("Invalid IP Address")

        if self.ssh_client is None:
            self.ssh_client = self.create_client()

        try:
            stdin, stdout, stderr = self.ssh_client.exec_command(cmd, timeout=timeout)
        except Exception:
            # Normally this is a timeout exception
            self.ssh_client = self.create_client()
            stdin, stdout, stderr = self.ssh_client.exec_command(cmd, timeout=timeout)

        out = None
        if not run_async:
            out = stdout.read().decode().strip()
            stderr.read().decode().strip()

        return out

    def download_remote_file(self, remote_src, local_dst):
        """
        Downloads a remote file to a local destination
        param: local_src: local file path source
        param: remote_dst: remote file path destination
        """
        if self.ssh_client is None:
            self.ssh_client = self.create_client()

        dirname = os.path.dirname(local_dst)
        if dirname and not os.path.exists(dirname):
            os.makedirs(dirname)

        ftp_client = self.ssh_client.open_sftp()
        ftp_client.get(remote_src, local_dst)
        ftp_client.close()

    def upload_local_file(self, local_src, remote_dst):
        """
        Upload a local file to a rempote destination
        param: local_src: local file path source
        param: remote_dst: remote file path destination
        """
        if self.ssh_client is None:
            self.ssh_client = self.create_client()

        ftp_client = self.ssh_client.open_sftp()
        ftp_client.put(local_src, remote_dst)
        ftp_client.close()

    def upload_multiple_local_files(self, file_list):
        """
        upload multiple files with the same sftp connection
        param: file_list: list of tuples [(local_src, remote_dst),]
        """
        if self.ssh_client is None:
            self.ssh_client = self.create_client()

        ftp_client = self.ssh_client.open_sftp()
        for local_src, remote_dst in file_list:
            ftp_client.put(local_src, remote_dst)
        ftp_client.close()

    def upload_data_to_file(self, data, remote_dst):
        """
        upload data to a remote file
        param: data: string data
        param: remote_dst: remote file path destination
        """
        if self.ssh_client is None:
            self.ssh_client = self.create_client()

        ftp_client = self.ssh_client.open_sftp()

        with ftp_client.open(remote_dst, "w") as f:
            f.write(data)

        ftp_client.close()



import logging
from collections import defaultdict
import logging.handlers
import os
import threading
from multiprocessing import Queue
from multiprocessing.managers import DictProxy
from queue import Empty
from traceback import TracebackException
from typing import Dict, List, Tuple
from flask import Flask, jsonify, request
from werkzeug.serving import make_server

from skyplane.chunk import ChunkRequest, ChunkState
from skyplane.gateway.chunk_store import ChunkStore
from skyplane.gateway.operators.gateway_receiver import GatewayReceiver
from skyplane.utils import logger


class GatewayDaemonAPI(threading.Thread):
    """
    API documentation:
    * GET /api/v1/status - returns status of API
    * GET /api/v1/servers - returns list of running servers
    * POST /api/v1/servers - starts a new server
    * DELETE /api/v1/servers/<int:port> - stops a server
    * GET /api/v1/chunk_requests - returns list of chunk requests (use {'state': '<state>'} to filter)
    * GET /api/v1/chunk_requests/<chunk_id> - returns chunk request
    * POST /api/v1/chunk_requests - adds a new chunk request
    * PUT /api/v1/chunk_requests/<chunk_id> - updates chunk request
    * GET /api/v1/chunk_status_log - returns list of chunk status log entries
    * POST /api/v1/upload_id_maps - post a json which contains mapping of region:bucket:key to upload id to each server
    """

    def __init__(
        self,
        chunk_store: ChunkStore,
        gateway_receiver: GatewayReceiver,
        error_event,
        error_queue: Queue,
        terminal_operators: Dict[str, List[str]],
        num_required_terminal: Dict[str, int],
        upload_id_map: DictProxy,
        host="0.0.0.0",
        port=8081,
    ):
        super().__init__()
        self.app = Flask("gateway_metadata_server")
        self.chunk_store = chunk_store
        self.gateway_receiver = gateway_receiver
        self.error_event = error_event
        self.error_queue = error_queue
        self.terminal_operators = terminal_operators
        self.num_required_terminal = num_required_terminal
        self.error_list: List[TracebackException] = []
        self.error_list_lock = threading.Lock()
        self.upload_id_map = upload_id_map

        # load routes
        self.register_global_routes(self.app)
        self.register_server_routes(self.app)
        self.register_request_routes(self.app)
        self.register_error_routes(self.app)
        self.register_socket_profiling_routes(self.app)

        # make server
        self.host = host
        self.port = port
        self.url = "http://{}:{}".format(host, port)

        # chunk status log
        self.state_update_lock = threading.Lock()
        self.chunk_status: Dict[int, str] = {}  # TODO: maintain as chunk_status_log is dumped
        self.chunk_requests: Dict[str, ChunkRequest] = {}
        self.sender_compressed_sizes: Dict[str, Tuple[int, int]] = {}  # TODO: maintain as chunks are completed
        self.chunk_status_log: List[Dict] = []
        self.chunk_completions = defaultdict(list)

        # socket profiles
        # TODO: actually fill these out
        self.sender_socket_profiles: List[Dict] = []
        self.sender_socket_profiles_lock = threading.Lock()
        self.receiver_socket_profiles: List[Dict] = []
        self.receiver_socket_profiles_lock = threading.Lock()

        logging.getLogger("werkzeug").setLevel(logging.DEBUG)
        self.server = make_server(host, port, self.app, threaded=True)

    def pull_chunk_status_queue(self, timeout=0.5):
        with self.state_update_lock:
            while True:
                try:
                    elem = self.chunk_store.chunk_status_queue.get(timeout=timeout)
                except Empty:
                    # print("[gateway_api] Chunk status queue empty, no more updates")
                    break

                handle = elem["handle"]
                state = elem["state"]
                chunk_id = elem["chunk_id"]
                partition = elem["partition"]
                chunk_file_path = self.chunk_store.get_chunk_file_path(elem["chunk_id"])

                if chunk_id not in self.chunk_status:
                    self.chunk_status[chunk_id] = ChunkState.registered.name

                if self.chunk_status[chunk_id] == ChunkState.complete.name:
                    assert not os.path.exists(chunk_file_path), f"Chunk path still exists even though completed {chunk_file_path}"
                    continue

                # if terminal operator, then mark a chunk completion
                if handle in self.terminal_operators[elem["partition"]] and state == ChunkState.complete.name:
                    self.chunk_completions[chunk_id].append(handle)

                # if all terminal operators complete, then mark chunk complete
                if (
                    self.chunk_status.get(chunk_id, None) != ChunkState.complete.name
                    and len(self.chunk_completions[chunk_id]) == self.num_required_terminal[partition]
                ):
                    # TODO: set this somewhere else
                    self.chunk_status[chunk_id] = ChunkState.complete.name

                    print(f"[gateway_api] chunk {chunk_id}: complete, all operators have uploaded {self.terminal_operators}")
                    # remove chunk file
                    if os.path.exists(chunk_file_path):
                        logging.info(f"[gateway_api] Removing chunk file {chunk_file_path}")
                        chunk_file_path.unlink()

                    # record compressed size
                    if "metadata" in elem and "compressed_size_bytes" in elem["metadata"] and "uncompressed_size_bytes" in elem["metadata"]:
                        self.sender_compressed_sizes[chunk_id] = (
                            elem["metadata"]["compressed_size_bytes"],
                            elem["metadata"]["uncompressed_size_bytes"],
                        )
                else:
                    if elem["state"] == ChunkState.complete.name:
                        print(
                            f"[gateway_api] After {handle}, chunk {chunk_id}, partition {partition}: not complete "
                            + f"operators {self.chunk_completions[chunk_id]} have uploaded "
                            + f"out of {self.terminal_operators}. "
                            + f"Required completitions = {self.num_required_terminal[partition]}"
                        )

                    # else:
                    #    print(f"[gateway_api] chunk {chunk_id}: after {handle} state = {elem['state']}")

                # only update chunk status log with terminal operators
                # otherwise, the client needs to filter chunk updates depending on whether the operator is terminal or not
                # this would require us to inform the client about the terminal operators, whcih seems annoying (though doable)
                # we can change this if we need to profile the chunk status progression through the DAG in detail
                if elem["state"] == ChunkState.complete.name:
                    if handle in self.terminal_operators[elem["partition"]]:
                        self.chunk_status_log.append(elem)
                else:
                    self.chunk_status_log.append(elem)

    def run(self):
        self.server.serve_forever()

    def shutdown(self):
        self.server.shutdown()

    def register_global_routes(self, app):
        @app.route("/", methods=["GET"])
        def get_index():
            return jsonify({"version": "v1"})

        @app.route("/api/v1", methods=["GET"])
        def get_v1_index():
            output = ""
            for rule in sorted(self.app.url_map.iter_rules(), key=lambda r: r.rule):
                if rule.endpoint != "static":
                    methods = set(m for m in rule.methods if m not in ["HEAD", "OPTIONS"])
                    output += f"<a href='{rule.rule}'>{rule.rule}</a>: {methods}<br>"
            return output

        @app.route("/api/v1/status", methods=["GET"])
        def get_status():
            return jsonify({"status": "ok"})

        @app.route("/api/v1/shutdown", methods=["POST"])
        def shutdown():
            self.shutdown()
            logger.error("Shutdown complete. Hard exit.")
            os._exit(1)

    def register_server_routes(self, app):
        @app.route("/api/v1/servers", methods=["GET"])
        def get_server_ports():
            return jsonify({"server_ports": self.gateway_receiver.server_ports})

        @app.route("/api/v1/servers", methods=["POST"])
        def add_server():
            new_port = self.gateway_receiver.start_server()
            return jsonify({"server_port": new_port})

        @app.route("/api/v1/servers/<int:port>", methods=["DELETE"])
        def remove_server(port: int):
            try:
                self.gateway_receiver.stop_server(port)
                return jsonify({"status": "ok"})
            except ValueError as e:
                return jsonify({"error": str(e)}), 400

    def register_request_routes(self, app):
        def make_chunk_req_payload(chunk_req: ChunkRequest):
            state = self.chunk_status[chunk_req.chunk.chunk_id]
            state_name = state if state is not None else "unknown"
            return {"req": chunk_req.as_dict(), "state": state_name}

        def get_chunk_reqs(state=None) -> Dict[int, Dict]:
            out = {}
            for chunk_id in list(self.chunk_status.keys()):
                chunk_state = self.chunk_status[chunk_id]
                if state is None or chunk_state == state:
                    chunk_req = self.chunk_requests[chunk_id]
                    out[chunk_id] = make_chunk_req_payload(chunk_req)
            return out

        def add_chunk_req(body, state):
            if isinstance(body, dict):
                chunk_req = ChunkRequest.from_dict(body)
                self.chunk_requests[chunk_req.chunk.chunk_id] = chunk_req
                qsize, succ = self.chunk_store.add_chunk_request(chunk_req, state)
                if not succ:
                    return 0, qsize, False
                return 1, qsize, True
            else:
                assert isinstance(body, list), f"Body must be list, got {type(body)}"
                added = 0
                for row in body:
                    chunk_req = ChunkRequest.from_dict(row)
                    self.chunk_requests[chunk_req.chunk.chunk_id] = chunk_req
                    qsize, succ = self.chunk_store.add_chunk_request(chunk_req, state)
                    if not succ:
                        return added, qsize, False
                    added += 1
                return added, qsize, True

        @app.route("/api/v1/chunk_requests", methods=["GET"])
        def get_chunk_requests():
            state_param = request.args.get("state")
            if state_param is not None:
                try:
                    state = ChunkState.from_str(state_param)
                except ValueError:
                    return jsonify({"error": "invalid state"}), 400
                return jsonify({"chunk_requests": get_chunk_reqs(state)})
            else:
                return jsonify({"chunk_requests": get_chunk_reqs()})

        @app.route("/api/v1/incomplete_chunk_requests", methods=["GET"])
        def get_incomplete_chunk_requests():
            return jsonify({"chunk_requests": {k: v for k, v in get_chunk_reqs().items() if v["state"] != "complete"}})

        # lookup chunk request given chunk worker_id
        @app.route("/api/v1/chunk_requests/<int:chunk_id>", methods=["GET"])
        def get_chunk_request(chunk_id: int):
            chunk_req = self.chunk_requests.get(chunk_id)
            if chunk_req:
                return jsonify({"chunk_requests": [make_chunk_req_payload(chunk_req)]})
            else:
                return jsonify({"error": f"Chunk {chunk_id} not found"}), 404

        # add a new chunk request with default state registered
        @app.route("/api/v1/chunk_requests", methods=["POST"])
        def add_chunk_request():
            print(f"[gateway_api] Recieved chunk request {request.json}")
            state_param = request.args.get("state", "registered")
            n_added, qsize, succ = add_chunk_req(request.json, ChunkState.from_str(state_param))
            # TODO: Add to chunk manager queue
            print(f"[gateway_api] Added {n_added} chunk requests to queue, size {qsize} success {succ}")
            return jsonify({"status": succ, "n_added": n_added, "qsize": qsize})

        # update chunk request
        @app.route("/api/v1/chunk_requests/<chunk_id>", methods=["PUT"])
        def update_chunk_request(chunk_id: str):
            chunk_req = self.chunk_requests.get(chunk_id)
            if chunk_req is None:
                return jsonify({"error": f"Chunk {chunk_id} not found"}), 404
            else:
                if "state" in request.args:
                    try:
                        state = ChunkState.from_str(request.args.get("state"))
                    except ValueError:
                        return jsonify({"error": "invalid state"}), 400
                    self.chunk_store.log_chunk_state(chunk_req, state)
                    return jsonify({"status": "ok"})
                else:
                    return jsonify({"error": "update not supported"}), 400

        # list chunk status log
        @app.route("/api/v1/chunk_status_log", methods=["GET"])
        def get_chunk_status_log():
            n_entries = len(self.chunk_status_log)
            status_log_copy = []  # copy to support concurrent access
            for i in range(n_entries):
                status_log_copy.append(self.chunk_status_log[i])
            return jsonify({"chunk_status_log": status_log_copy})

        # post the upload ids mapping
        @app.route("/api/v1/upload_id_maps", methods=["POST"])
        def update_upload_ids_mapping():
            # TODO: beware that this assumes that only a single thread on the client is making requests
            # if concurrent calls are made, this needs to be processed as chunk requests are
            logging.debug(f"[gateway_api] Recieved id mapping request {request.json}")
            # update upload id mapping
            upload_ids = request.json
            for region_tag in upload_ids.keys():
                for key, upload_id in upload_ids[region_tag].items():
                    self.upload_id_map[region_tag + key] = upload_id
            return jsonify({"status": "ok"})

    def register_error_routes(self, app):
        @app.route("/api/v1/errors", methods=["GET"])
        def get_errors():
            with self.error_list_lock:
                while True:
                    try:
                        elem = self.error_queue.get_nowait()
                        self.error_list.append(elem)
                    except Empty:
                        break
                # convert TracebackException to list
                error_list_str = [str(e) for e in self.error_list]
                return jsonify({"errors": error_list_str})

    def register_socket_profiling_routes(self, app):
        @app.route("/api/v1/profile/socket/receiver", methods=["GET"])
        def get_receiver_socket_profiles():
            with self.receiver_socket_profiles_lock:
                while True:
                    try:
                        elem = self.gateway_receiver.socket_profiler_event_queue.get_nowait()
                        self.receiver_socket_profiles.append(elem)
                    except Empty:
                        break
                return jsonify({"socket_profiles": self.receiver_socket_profiles})

        @app.route("/api/v1/profile/compression", methods=["GET"])
        def get_receiver_compression_profile():
            total_size_compressed_bytes, total_size_uncompressed_bytes = 0, 0
            for _, (compressed_size, uncompressed_size) in self.sender_compressed_sizes.items():
                total_size_compressed_bytes += compressed_size
                total_size_uncompressed_bytes += uncompressed_size
            return jsonify(
                {
                    "compressed_bytes_sent": total_size_compressed_bytes,
                    "uncompressed_bytes_sent": total_size_uncompressed_bytes,
                    "compression_ratio": total_size_uncompressed_bytes / total_size_compressed_bytes
                    if total_size_compressed_bytes > 0
                    else 0,
                }
            )



# (C) Copyright Samsung SDS. 2023

#

# Licensed under the Apache License, Version 2.0 (the "License");

# you may not use this file except in compliance with the License.

# You may obtain a copy of the License at

#

#     http://www.apache.org/licenses/LICENSE-2.0

#

# Unless required by applicable law or agreed to in writing, software

# distributed under the License is distributed on an "AS IS" BASIS,

# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.

# See the License for the specific language governing permissions and

# limitations under the License.
import json
from typing import List, Optional
from skyplane.config_paths import config_path, scp_config_path, scp_quota_path

from skyplane.compute.server import key_root
from skyplane.config import SkyplaneConfig
from skyplane.compute.scp.scp_utils import SCPClient
from skyplane.compute.scp.scp_utils import CREDENTIALS_PATH


class SCPAuthentication:
    def __init__(self, config: Optional[SkyplaneConfig] = None):
        if not config is None:
            self.config = config
        else:
            self.config = SkyplaneConfig.load_config(config_path)

        self._scp_access_key = self.config.scp_access_key
        self._scp_secret_key = self.config.scp_secret_key
        self._scp_project_id = self.config.scp_project_id

    def _get_scp_vm_quota(self, region):
        """scp vm quota - no restirction on vcpu, but if no enough quota, scp will return error"""
        result_list = []
        for region_name in region:
            region_dict = {"on_demand_standard_vcpus": 500, "service_zone_name": region_name}
            result_list.append(region_dict)

        return result_list

    def save_region_config(self, config: SkyplaneConfig):
        if self.config.scp_enabled == False:
            self.clear_region_config()
            return
        region_list = self.get_region_config()

        with open(scp_config_path, "w") as f:
            for region in region_list:
                f.write(region + "\n")

        quota_infos = self._get_scp_vm_quota(region_list)
        with scp_quota_path.open("w") as f:
            f.write(json.dumps(quota_infos, indent=2))

    @staticmethod
    def get_zones() -> dict:
        scp_client = SCPClient()
        url = f"/project/v3/projects/{scp_client.project_id}/zones"

        response = scp_client._get(url)
        return response

    @staticmethod
    def get_region_config() -> List[str]:
        zones = SCPAuthentication.get_zones()
        # print(zones)
        service_zone_locations = [zone["serviceZoneName"] for zone in zones]
        return service_zone_locations

    def clear_region_config():
        with scp_config_path.open("w") as f:
            f.write("")
        with scp_quota_path.open("w") as f:
            f.write("")

    def enabled(self):
        return self.config.scp_enabled

    def get_zone_location(self, zone_name):
        zones = self.get_zones()
        for zone in zones:
            if zone["serviceZoneName"] == zone_name:
                return zone["serviceZoneLocation"]
        return None

    @property
    def credential_path(self):
        credential_path = key_root / "scp" / "scp_credential"
        return credential_path

    @property
    def scp_credential_path(self):
        return CREDENTIALS_PATH

    @property
    def scp_access_key(self):
        if self._scp_access_key is None:
            self._scp_access_key, self._scp_secret_key, self._scp_project_id = self.infer_credentials()
        return self._scp_access_key

    @property
    def scp_secret_key(self):
        if self._scp_secret_key is None:
            self._scp_access_key, self._scp_secret_key, self._scp_project_id = self.infer_credentials()
        return self._scp_secret_key

    @property
    def scp_project_id(self):
        if self._scp_project_id is None:
            self._scp_access_key, self._scp_secret_key, self._scp_project_id = self.infer_credentials()
        return self._scp_project_id

    def infer_credentials(self):
        scp_client = SCPClient()
        return scp_client.access_key, scp_client.secret_key, scp_client.project_id
